/* Copyright (c) 2009 The Regents of the Signet Group.
 *  All rights reserved.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose, without fee, and without written
 *  agreement is hereby granted, provided that the above copyright
 *  notice, the (updated) modification history and the author appear in
 *  all copies of this source code.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS `AS IS'
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS
 *  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, LOSS OF USE, DATA,
 *  OR PROFITS) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *  THE POSSIBILITY OF SUCH DAMAGE.
 *  
 *  Authors:
 *  Angelo P. Castellani <castellani@dei.unipd.it>
 *  Mattia Gheda <ghedamat@dei.unipd.it>
 */

/*
 * CoAP header
 *      0                   1                   2                   3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |Ver| T |  OC   |      Code     |        Transaction ID         |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 */

#ifdef i386
typedef struct  {
	nx_uint8_t oc:4;
	nx_uint8_t t:2;
	nx_uint8_t ver:2;
	nx_uint8_t code;
	nx_uint16_t tid;
} coap_header_t;
#else
typedef nx_struct  {
	nx_uint8_t ver:2;
	nx_uint8_t t:2;
	nx_uint8_t oc:4;
	nx_uint8_t code;
	nx_uint16_t tid;
} coap_header_t __attribute__ ((__packed__));
#endif

#ifdef i386
typedef struct {
	nx_uint8_t len:4;
	nx_uint8_t od:4; //option delta
} coap_option_header_t ;
#else
typedef nx_struct {
	nx_uint8_t od:4; //option delta
	nx_uint8_t len:4;
} coap_option_header_t __attribute__ ((__packed__));

#endif

/*
 * +------+-----+---------------+--------------+--------+--------------+
   | Type | C/E | Name          | Data type    | Length | Default      |
   +------+-----+---------------+--------------+--------+--------------+
   | 0    | -   | Reserved      | -            | -      | -            |
   | 1    | C   | Content-type  | 8-bit        | 1 B    | 0            |
   |      |     |               | unsigned     |        | (text/plain) |
   |      |     |               | integer      |        |              |
   | 2    | E   | Max-age       | Variable     | 1-4 B  | 60 seconds   |
   |      |     |               | length       |        |              |
   |      |     |               | unsigned     |        |              |
   |      |     |               | integer      |        |              |
   | 3    | C   | Uri-Scheme    | String       | 1-270  | "coap"       |
   |      |     |               |              | B      |              |
   | 4    | E   | Etag          | Sequence of  | 1-4 B  | -            |
   |      |     |               | bytes        |        |              |
   | 5    | C   | Uri-Authority | String       | 1-270  | ""           |
   |      |     |               |              | B      |              |
   | 6    | E   | Location      | String       | 1-270  | -            |
   |      |     |               |              | B      |              |
   | 7    | -   | Reserved      | -            | -      | -            |
   | 9    | C   | Uri-Path      | String       | 1-270  | ""           |
   |      |     |               |              | B      |              |
   +------+-----+---------------+--------------+--------+--------------+

                          Table 2: Option headers 
 */
 
enum coap_option_types {
 	COAP_OPTION_CONTENT_TYPE = 1,
 	COAP_OPTION_MAX_AGE,
 	COAP_OPTION_URI_SCHEME,
 	COAP_OPTION_ETAG,
 	COAP_OPTION_URI_AUTHORITY,
 	COAP_OPTION_LOCATION,
 	COAP_OPTION_URI_PATH = 9,
 	COAP_OPTION_BLOCK = 13 // BETA
 	
};

enum coap_v_values {
	COAP_VERSION = 1
};

enum coap_t_values {
	COAP_T_CON = 0,
	COAP_T_NON,
	COAP_T_ACK,
	COAP_T_RST
};

enum coap_method_values {
	COAP_METHOD_GET = 1,
	COAP_METHOD_POST,
	COAP_METHOD_PUT,
	COAP_METHOD_DELETE
	
};

enum coap_msgtype_values {
	COAP_MSGTYPE_REQUEST = 0,
	COAP_MSGTYPE_RESPONSE
};

#warning TODO add other codes see Sec. 11.1 enum per gli status

enum coap_status_values {
        COAP_STATUS_OK = 80,
        COAP_STATUS_OK_NO_CONTENT = 81,
        COAP_STATUS_OK_BUFFERED_DATA = 82,
        COAP_STATUS_BAD_REQUEST = 160,
        COAP_STATUS_NOT_FOUND = 164,
        COAP_STATUS_FAILURE = 200
};

// content type possibili 
#warning TODO: edit this according to draft?
enum coap_contenttype_values {	
	COAP_CONTENTTYPE_EXI = 47,
	COAP_CONTENTTYPE_PLAIN = 0
};

void parse_packet( coap_packet_t* pkt, coap_msg_t* msg) {
	coap_char_t *p,*u; //pkt pointer
	uint8_t oc; //option counter
	uint8_t lastdelta; //last option delta
	coap_header_t* h; //msg header static
	
	p = (coap_char_t*)pkt->buf;
	h = (coap_header_t*)pkt->buf;	

	/*
	 * PARSE BASE HEADER
	 */
	oc = h->oc; // get option count
	
	msg->ack = FALSE;
	msg->method = 0;
	msg->status = 0;
	switch(h->t){
		case COAP_T_CON:
			msg->ack = TRUE; // SOLO QUESTO CASO?
		case COAP_T_NON:
			pkt->msgtype = COAP_MSGTYPE_REQUEST;
			msg->method = h->code;
		break;
		case COAP_T_ACK:
		case COAP_T_RST:
			pkt->msgtype = COAP_MSGTYPE_RESPONSE;
			msg->status = h->code;
		break;
	}	

#ifdef i386
	pkt->tid = ntohs(h->tid);
#else
	pkt->tid = h->tid; // XXX USE GET16_T here?
#endif
	
	/*
	 * PARSE VARIABLE OPTIONS
	 */
	// setting default op for content type (see 3.2.1)
	msg->content->format = COAP_CONTENTTYPE_PLAIN;

	p+=sizeof(coap_header_t);

	u = msg->absuri->uri;
	*(u++) = '/';

	lastdelta =0; 
	while (oc){ // option count > 0
		coap_char_t delta;
		coap_char_t len;
		
		delta = ((coap_option_header_t* )p)->od + lastdelta;
		len = ((coap_option_header_t* )p)->len;
		p+=sizeof(coap_option_header_t); // p++;
		if (len == 0xf) // len = 15
			len += *(coap_char_t*)p++;
		
		switch(delta){
			case COAP_OPTION_CONTENT_TYPE:
				msg->content->format = *(coap_char_t*)p++;
			break;
			
			case COAP_OPTION_URI_PATH:{ // see 3.2.4 ! dubbi?
				while (len--) { *(u++) = *(p++); }
				*(u++) ='\0';
			}
			break;
			
#ifdef COAP_BLOCK_OPTION
			case COAP_OPTION_BLOCK:{
				uint8_t tmp[4];
				uint32_t ob = 0;
				msg->chunked = TRUE;
				
				memcpy(&tmp,p,len);
				ob = *(uint32_t*)&tmp;

				msg->szx = ob & 0x7;
				msg->mb = (ob & 0x8) >> 3;
				msg->bn = ob >> 4;
				
				p+=len;
			}
			break;
#endif		
			default:
				p+=len;
			break;
		}
		
		lastdelta = delta;
		oc--;
	}
	
	msg->content->data = p;
	msg->content->len = pkt->len - (p-pkt->buf);
}

void write_packet( coap_packet_t* pkt, coap_msg_t* msg ) {
	// ho in pkt la cosa da scrivere in msg i dati;
	coap_char_t* p; //pkt pointer
	uint8_t oc = 0; //option counter
	uint8_t lastdelta = 0; //last option delta
	coap_header_t* h; //msg header static
	
	p = (coap_char_t*)pkt->buf;
	h = (coap_header_t*)pkt->buf;
	
	/*
	 * Write base header and opts
	 */
	p+=sizeof(coap_header_t);
	 
	h->ver = 1;
#ifdef i386
	h->tid = htons(pkt->tid);
#else
	h->tid = pkt->tid;
#endif

	/*
	 * set CONTENT TYPE OPTION : MANDATORY
	 */  

	if(msg->content){
		((coap_option_header_t* )p)->od = COAP_OPTION_CONTENT_TYPE;
		((coap_option_header_t* )p)->len = 1;
		p+=sizeof(coap_option_header_t); // p++;
		*(coap_char_t*)p++ = msg->content->format;
		lastdelta=COAP_OPTION_CONTENT_TYPE;
		oc++;
	}
		//TODO ADD THIS Later *(coap_char_t*)p++ = COAP_CONTENTTYPE_PLAIN;
	
	
	switch(pkt->msgtype){
		case COAP_MSGTYPE_REQUEST:{
			coap_char_t len;
			coap_char_t* u = msg->absuri->uri;
			
			/*
			 * set BASE Header
			 */
			h->t = (msg->ack == TRUE ? COAP_T_CON : COAP_T_NON);
			h->code = msg->method;	
			
			/*
			 * set URI OPTION
			 */
			if (*u == '/')
				u++;

			len = strlen(u); // la devo conoscere prima ( il campo len putroppo sono 4 o 12 bits)
			if (len) {
				oc++;
				((coap_option_header_t* )p)->od = COAP_OPTION_URI_PATH-lastdelta;
				if (len <0xf){
					((coap_option_header_t* )p)->len = len;
					p+=sizeof(coap_option_header_t); // p++;
				}else{
					((coap_option_header_t* )p)->len = 0xf;
					p+=sizeof(coap_option_header_t); // p++;
					*(coap_char_t*)p++ = len - 0xf;
				}
				lastdelta = COAP_OPTION_URI_PATH; // usless per ora, servira forse con piu' opzioni
					
				while (*u) { *(p++) = *(u++);}
			}
		}
		break;
		case COAP_MSGTYPE_RESPONSE:
			h->t = COAP_T_ACK; // TODO: aggiungere RST?
			h->code = msg->status;
		break;
	}
	
#ifdef COAP_BLOCK_OPTION
	if (msg->chunked) {
		uint8_t len = 1;
		coap_option_header_t* block = (coap_option_header_t*) p;
		p += sizeof(coap_option_header_t);
		
		if (msg->bn & 0x000ff000)
			len++, *(p++) = (0x000ff000 & msg->bn) >> 12;

		if (msg->bn & 0x00000ff0)
			len++, *(p++) = (0x00000ff0 & msg->bn) >> 4;

		*p = (0x0000000f & msg->bn)<<4;
		
		if (msg->mb)
			*p |= 0x8;
		
		*(p++) |=  msg->szx & 0x7;
		
		block->od = COAP_OPTION_BLOCK-lastdelta;
		block->len = len;

		oc++;
	}
#endif
	
	h->oc = oc;	
	/*
	 * set the payload
	 */
	if (msg->content) {
		memcpy(p, msg->content->data, msg->content->len);
		p += msg->content->len;
	}

	pkt->len = p - pkt->buf;
	 
}


